#!/usr/bin/env python

# Copyright 2014 Keith Clark, Peter Robinson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import pedroclient

import threading
import Queue 

import Tkinter as tk
import argparse

DEFAULT_ENTRIES = "5"
SEARCH_TEXT = "Received new percepts"
DELAY = 500

message_queue = Queue.Queue()

class MessageThread(threading.Thread):
    def __init__(self, client):
        self.running = True
        self.client = client
        threading.Thread.__init__(self)
        self.daemon = True
        
    def run(self):
        while self.running:
            
            p2pmsg = self.client.get_term()[0]
            message = str(p2pmsg.args[2])[1:-1]
            tr_addr = p2pmsg.args[1].args[0].args[1].val
            message_queue.put(str(tr_addr) +': '+message+'\n')
            
    def stop(self):
        self.running = False




class LoggerApp(object):
    def __init__(self, logger_name, host, master):
        master.title("Logger : " + logger_name)
        master.protocol("WM_DELETE_WINDOW", self.quit)
        self.master = master
        master.geometry("470x500")
        self.text = tk.Text(master)
        self.text.pack(fill=tk.BOTH, expand=1)
        self.master.bind('<<msg>>', self.process_message)
        snapshot_frame = tk.Frame(master)
        snapshot_frame.pack()
        snapshot_label = tk.Label(snapshot_frame,text="Number of Entries: ")
        snapshot_label.pack(side=tk.LEFT)
        var = tk.StringVar(root)
        var.set(DEFAULT_ENTRIES)
        self.snapshot_scale = \
            tk.Spinbox(snapshot_frame, from_=1,to=20,width = 5, 
                       textvariable=var)
        self.snapshot_scale.pack(side=tk.LEFT)
        snapshot_button = tk.Button(snapshot_frame, text="Snap", 
                                    command=self.snap)
        snapshot_button.pack(side=tk.LEFT)
        self.client = pedroclient.PedroClient(host)
        if self.client.register(logger_name):
            self.text.insert(tk.END, "Logger ready\n\n", "blue")
            self.text.see(tk.END)
        else:
            self.text.insert(tk.END, "Cannot register "+ logger_name + " with Pedro\n\n", "red")
            self.text.see(tk.END)
        self.message_thread = MessageThread(self.client)
        self.message_thread.start()
        self.text.tag_config("blue", foreground="blue")
        self.text.tag_config("red", foreground="red")
        self.do_process = True

    def process_message(self):
        progress = False
        while self.do_process and not message_queue.empty():
            progress = True
            msg = message_queue.get()
            self.text.insert(tk.END, msg)
        if progress: 
            self.text.see(tk.END)
        self.master.after(DELAY,self.process_message)
        
    def snap(self):
        self.do_process = False
        current_end = tk.END
        try:
            num = int(self.snapshot_scale.get())
        except:
            num = int(DEFAULT_ENTRIES)
        pos = current_end
        for i in range(num):
            pos = self.text.search(SEARCH_TEXT, pos, backwards=True)
            if pos == "":
                pos = "1.0"
                break
        SnapWindow(self.text.get(pos, current_end))
        self.do_process = True
 
    def quit(self):
        self.message_thread.stop()
        self.master.destroy()

class SnapWindow(tk.Toplevel):
    def __init__(self, text) :
        tk.Toplevel.__init__(self)
        self.title("Snapshot")
        self.geometry("470x450")
        self.text = tk.Text(self)
        self.text.pack(fill=tk.BOTH, expand=1)
        self.text.insert(tk.END, text)
        self.text.see(tk.END)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("logger_name", help="the process name of this logger")
    parser.add_argument("-N", dest="pedro", default='localhost', 
                      help="the address of the Pedro server")
    args = parser.parse_args()
    root = tk.Tk()
    app = LoggerApp(args.logger_name, args.pedro, root)
    root.after(DELAY, app.process_message)
    root.mainloop()


