#!/usr/bin/env python

# This file gives a simple example of an interface between ROS and Pedro
# The intention is that this program "sits between" a robot running ROS
# and a TR agent used to control the robot.
#
# For this simple example the commands coming from the agent are converted into
# Twist messages published on the '/base/cmd_vel' topic.

import rospy
#this robot uses Twist for command messages
from geometry_msgs.msg import Twist
from std_msgs.msg import String

import pedroclient

import threading

class MessageThread(threading.Thread):
    def __init__(self, parent):
        self.running = True
        self.parent = parent
        threading.Thread.__init__(self)
        self.daemon = True
        # for the emu robot the controls are via the topic '/base/cmd_vel'
        self.pub = rospy.Publisher('/base/cmd_vel', Twist, queue_size=10)
        rospy.init_node('ros_pedro', anonymous=True)
        self.r = rospy.Rate(10)
        
    def run(self):
        while self.running:
            p2pmsg = self.parent.client.get_term()[0]
            # get the message
            message = p2pmsg.args[2]
            print str(message)
            if str(message) == 'initialise_':
                # get the sender address
                percepts_addr = p2pmsg.args[1]
                #print percepts_addr
                self.parent.set_client(percepts_addr)
                # for this robot we will ignore the 'initialise_' message
                continue
            # Twist uses linear and angular velocities so we will use the
            # command move(linear_velocity, angular_velocity)
            #    linear_velocity > 0 => move forward
            #    angular_velocity > 0 => turn left
            # (both numbers)
            # The agent messages will be of the form
            # start_(move(lin_vel, ang_vel)) - starts a move
            # mod_(move(lin_vel, ang_vel)) - modify the velocities
            # stop_(move(lin_vel, ang_vel)) - stop moving
            if message.get_type() != pedroclient.PObject.structtype:
                continue
            functor = message.functor
            if functor.get_type() != pedroclient.PObject.atomtype:
                continue
            cmd_type = functor.val
            cmd = message.args[0]
            if cmd.get_type() != pedroclient.PObject.structtype:
                continue
            if cmd_type == 'stop_':
                if cmd.functor.val == 'move' and cmd.arity() == 2:
                    twist = Twist();
                    twist.linear.x = 0
                    twist.angular.z = 0
                    self.pub.publish(twist)
                    self.r.sleep()
            elif cmd_type in ['start_', 'mod_']:
                if cmd.functor.val == 'move' and cmd.arity() == 2:
                    twist = Twist();
                    twist.linear.x = cmd.args[0].val
                    twist.angular.z = cmd.args[1].val
                    self.pub.publish(twist)
                    self.r.sleep()
            
    def stop(self):
        self.running = False

class RosPedro(object):
    def __init__(self):
        self.tr_client_addr = None
        self.client = pedroclient.PedroClient("localhost")
        self.client.register("ros_pedro")
        message_thread = MessageThread(self)
        message_thread.start()

    def set_client(self, addr):
        self.tr_client_addr = addr

    def percepts_callback(self, percepts):
        # A percepts message has arrived from the robot via ROS
        # In this case it is simply passed on to the agent as is
        # If some conversion is required then you might need some
        # Python string processing to convert into a form suitable
        # as a Pedro message for the agent
        print str(percepts)
        if self.client.p2p(self.tr_client_addr, str(percepts.data)) == 0:
            print "Error", percepts.data

rp = RosPedro()
rospy.Subscriber("percepts", String, rp.percepts_callback)
rospy.spin()
