#!/usr/bin/python

# Copyright 2014 Keith Clark, Peter Robinson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys,os
import wx
import wx.lib
import wx.lib.newevent
import threading
import random
import math
import pedroclient
import time
USING_SOUND = False
# In order to play the sounds you need pygame installed.
try:
    from pygame import mixer
    USING_SOUND = True
except:
    pass

TWO_TASK = False

NUM_ROBOTS = 2

black = 0, 0, 0
grey = 100,100,100
red = 255, 0, 0
blue = 0, 0, 255

ROBOT1_COLOUR = '#0000AA'
ROBOT2_COLOUR = '#b000b0'
BOTTLE_COLOUR = '#00AA00'
ROBOT_COLOUR = '#FFFF00'
CAMERA_ARC_COLOUR = '#C8C800'
GREY = '#AAAAAA'

FONT_SIZE = 20
FONT_NAME = 'Courier'

TING_FILE = 'tingLoud.wav'
SWOOSH_FILE = 'SwooshLoud.wav'

ROBOT1_COLOUR_TRIPLE = (int(ROBOT1_COLOUR[1:3], 16),
                        int(ROBOT1_COLOUR[3:5], 16),
                        int(ROBOT1_COLOUR[5:7], 16))
ROBOT2_COLOUR_TRIPLE = (int(ROBOT2_COLOUR[1:3], 16),
                        int(ROBOT2_COLOUR[3:5], 16),
                        int(ROBOT2_COLOUR[5:7], 16))

DISPLAY_TEXT = "Total: {0} Last Msg: {1}"

TWO_DISPLAY_TEXT = "Total: {0}"

ONE_DISPLAY_TEXT = "Total: {0}"


CAMERA_ANGLE = math.pi/4   # the camera vision angle
HC_ANGLE = CAMERA_ANGLE/2

RADAR_ANGLES = [0, math.pi/2, 3*math.pi/8, math.pi/4,
                -math.pi/2, -3*math.pi/8, -math.pi/4]
RADAR_NAMES = ['c', 'r3','r2','r1','l3','l2','l1']


WIDTH = 1200           # screen width
HEIGHT = 760          # screen height
TEXT_AREA_HEIGHT = 60

BOTTLE_RADIUS = 5

DROP_RADIUS = 50
DROP_POSITION = WIDTH/2, HEIGHT/2

ROBOT_RADIUS = 20

ROBOT_ID_RADIUS = 8

CLOSE_DIST = 40

NEAR_DIST = 80

MESSAGE_TIME = -1   ### 2000  # in miliseconds
DROP_TIME = 600  # in miliseconds

# square of distance between robot centres
SQ_R2R = (2*ROBOT_RADIUS)**2

# square of dist between robot and bottle centres 
SQ_R2C = (ROBOT_RADIUS + BOTTLE_RADIUS)**2 

# square of dist between robot and drop centres 
SQ_R2D = (ROBOT_RADIUS + DROP_RADIUS)**2 

# how close is near
CLOSE = 40

# square of 'close distance' between robots
C_SQ_R2R = (2*ROBOT_RADIUS + CLOSE)**2

# square of close dist between robot and bottle 
C_SQ_R2C = (ROBOT_RADIUS + BOTTLE_RADIUS + CLOSE)**2 

# square of dist between robot and drop 
C_SQ_R2D = (ROBOT_RADIUS + DROP_RADIUS + CLOSE)**2 

# close distance to wall
C_W = ROBOT_RADIUS + CLOSE

SQ_CW = C_W**2

# number of time steps for veering
VEER_STEPS = 80

TIMER_ID = 100
MESSAGE_TIMER_IDS = [0,101,102]
DROP_TIMER_IDS = [0,103,104]

SPEED_FACTOR = 2
RATE_FACTOR = 2

def dist2name(dist):
    if dist == 1000000:
        return 'not_see'
    if dist < CLOSE_DIST:
        return 'close_to'
    elif dist < NEAR_DIST:
        return 'near'
    else:
        return 'see'
 
def get_dir(x, y):
    """Return the direction of the vector (x,y) as
    a result in [0,2*pi]
    """
    if abs(x) < 1.0e-20:
        if y > 0:
            odir = math.pi/2
        else:
            odir = 3*math.pi/2
    else:
        odir = math.atan(abs(float(y)/x))
        if x < 0:
            odir = math.pi - odir
        if y < 0:
            odir = 2*math.pi - odir
    return odir

def normalize_angle(angle):
    while angle > math.pi:
        angle -= 2*math.pi
    while angle < -math.pi:
        angle += 2*math.pi
    return angle

def get_rel_dir(x,y,rdir):
    """Return the direction of the vector (x,y) relative to the direction rdir
    The result is in [-pi, pi]
    """
    return normalize_angle(get_dir(x,y) - rdir)
    
def get_rel_xy(obj1, obj2, rdir):
    """Get the x,y cords of obj2 relative to obj1 (using dir as x axis)"""
    x1 = obj2.x - obj1.x
    y1 = obj2.y - obj1.y
    s = math.sin(rdir)
    c = math.cos(rdir)
    return (c*x1 + s*y1, -s*x1 + c*y1)

def get_closest_to(obj1, obj2, rdir):
    """Get the closest distance from obj1 to obj2 in rdir. obj2 
    return 1000000 if not in overlap in rdir direction
    """
    x,y = get_rel_xy(obj1, obj2, rdir)
    r = obj2.radius
    sdiff = r**2 - y**2
    if (sdiff < 0.1) or (x < 0):
        return 1000000
    return x - math.sqrt(sdiff)

def dist_between(obj1, obj2):
    d1 = math.sqrt((obj1.x - obj2.x)**2 + (obj1.y - obj2.y)**2)
    return d1 - obj1.radius - obj2.radius

def dir2text(dist):
    if dist == 1:
        return "dead_centre"
    elif dist == 2:
        return "centre"
    elif dist == 4:
        return "right"
    else:
        return "left"

def approx_dist(n):
    if n < 2:
        return 0
    return 2**(int(round(math.log(n)/math.log(2))))

class WallPoint:
    def __init__(self, x, y):
        self.x = float(x)
        self.y = float(y)

corner0 = WallPoint(0,0)
corner1 = WallPoint(WIDTH,0)
corner2 = WallPoint(WIDTH,HEIGHT)
corner3 = WallPoint(0,HEIGHT)

def x_inter(x0, y0, x1, y1):
    if (y0 > 0 and y1 < 0) or (y0 < 0 and y1 > 0):
        d = x0 - y0*(x1 - x0)/(y1 - y0)
        if d > 0:
            return d
    return 1000000

def get_closest_wall(obj1, rdir):
    dist = 100000
    x0, y0 = get_rel_xy(obj1, corner0, rdir)
    x1, y1 = get_rel_xy(obj1, corner1, rdir)
    x2, y2 = get_rel_xy(obj1, corner2, rdir)
    x3, y3 = get_rel_xy(obj1, corner3, rdir)
    return min([x_inter(x0, y0, x1, y1), x_inter(x1, y1, x2, y2), 
                x_inter(x2, y2, x3, y3), x_inter(x3, y3, x0, y0)])
        
    


def see_dir(obj1, obj2,  angle = None):
    """ Direction from obj1 (robot) to obj2    
    0  - not seen
    1  - dead_centre
    2  - centre
    4  - right
    8  - left
    """
    if angle is None:
        angle = get_rel_dir(obj2.x - obj1.x, obj2.y - obj1.y, obj1.dir)
    offset = get_dir_edges(obj1.x, obj1.y, obj2.x, obj2.y, obj2.radius)
    left_angle = normalize_angle(angle - offset)
    right_angle = normalize_angle(angle + offset)
    if right_angle < -HC_ANGLE or left_angle > HC_ANGLE:
        return (0, left_angle, right_angle) 
    if -HC_ANGLE/8 < (left_angle + right_angle) < HC_ANGLE/8:
        return (1, left_angle, right_angle)  
    if -HC_ANGLE/2 < (left_angle + right_angle) < HC_ANGLE/2:
        return (2, left_angle, right_angle)  
    elif  (left_angle + right_angle) <=  -HC_ANGLE/2 and \
            right_angle >  -HC_ANGLE:
        return (8, left_angle, right_angle) 
    elif (left_angle + right_angle) >=   HC_ANGLE/2 and \
            left_angle < HC_ANGLE:
        return (4, left_angle, right_angle) 
    else:
        return (0, left_angle, right_angle) 

def get_dir_edges(x1, y1, x2, y2, radius):
    """Return the direction in [0,pi] of the right edge
    of a circular object of given radius whose centre is at
    (x2, y2) from the point (x1,y1)
    """
    hyp = math.sqrt((x1 - x2)**2 + (y1 - y2)**2)
    if radius > hyp: return HC_ANGLE
    return math.asin(radius/hyp)

def transform(shiftx, shifty, theta, pts):
    """Used for shift and rotation of the polygons representing the 
    robot wheels"""
    cos_theta = math.cos(theta)
    sin_theta = math.sin(theta)
    for i in range(len(pts)):
        x,y = pts[i]
        newx = shiftx + x*cos_theta - y*sin_theta
        newy = shifty + x*sin_theta + y*cos_theta
        pts[i] = (newx,newy)


# Create an event for notifications
(NotificationEvent, EVT_NOTIFICATION_EVENT) = wx.lib.newevent.NewEvent()


class GetNotifications(threading.Thread):
    def __init__( self, client, panel):
        self.client = client
        self.panel = panel
        self.running = True
        threading.Thread.__init__(self)

    def run(self):
        """Run as thread - post an event when notifications arrive"""
    
        while (self.running):
            try:
                msg = self.client.get_term()[0]
                evt = NotificationEvent(notification=msg)
                wx.PostEvent(self.panel, evt)
            except:
                break

    def Stop(self):
        self.running = False
    

class Drop:
    """A drop class so teh drop bottle be treated like a bottle object."""
    def __init__(self):
        self.x, self.y = DROP_POSITION
        self.radius = DROP_RADIUS

class Robot:
    def __init__(self, env, ID, x, y):
        self.x,self.y = x, y
        self.dir = 0
        self.holding = False
        self.last_holding = True
        self.gripper_open = True
        self.last_gripper_open = False
        self.moving = False
        self.percept_update = False
        self.turning = 0
        self.touching = 0
        self.saved_touching = 0
        self.seeing_drop = 0
        self.at_drop = False
        self.close_drop = False
        self.close_bottle = False
        self.close_robot = False
        self.speed = 1.5
        self.veer_steps = 0
        self.veering = False
        self.env = env
        self.ID = ID
        self.drop = Drop()
        self.over = False
        self.queued_messages = []
        self.radius = ROBOT_RADIUS
        self.saved_robot_see_dist = -1
        self.saved_drop_see_dist = -1
        self.saved_bottle_see_dist = -1
        self.saved_radar_right = -1
        self.saved_radar_left = -1
        self.saved_radar_centre = -1
        self.stopped = False
        self.message = ''
        self.dropped_can = None
        self.started = False
        self.message_timer = wx.Timer(self.env.parent, MESSAGE_TIMER_IDS[ID])
        wx.EVT_TIMER(self.env.parent, MESSAGE_TIMER_IDS[ID], 
                     self.delete_message)
        self.drop_timer = wx.Timer(self.env.parent, DROP_TIMER_IDS[ID])
        wx.EVT_TIMER(self.env.parent, DROP_TIMER_IDS[ID], 
                     self.delete_drop)
        self.count = 0

    def delete_message(self, evt):
        self.message = ''

    def delete_drop(self, env):
        if self.dropped_can is not None:
            if USING_SOUND:
                self.env.parent.swoosh_sound.play()
            #time.sleep(0.4)
            self.env.removeBottleID(self.dropped_can)
            self.dropped_can = None

    def add_message(self, msg):
        if TWO_TASK:
            self.queued_messages.append('msg(robot{0}, {1})'.format(self.ID, 
                                                                    msg))
        else:
            self.queued_messages.append(msg)

    def flipStopped(self):
        if self.stopped:
            self.add_message('set(simulation_stopped)')
        else:
            self.add_message('unset(simulation_stopped)')
        self.send_messages()
        self.stopped = not self.stopped

    def send_msg(self, msg):
        #print self.ID, msg
        self.env.send_msg(self.ID, msg)
        
    def send_messages(self):
        if self.queued_messages:
            message = '['
            for m in self.queued_messages:
                message += m + ','
            message = message[:-1]+']'
            self.send_msg(message)
            self.queued_messages = []
 
    def check_seeing_drop(self, see_robots):
        """Check if seeing drop - send off percept.
        see_robots is a list of  (ID, left, right) where left and right are the
        relative angles to the robot"""
        see = 0
        # angle to the centre of the drop
        angle = get_rel_dir(self.drop.x - self.x, self.drop.y - self.y, 
                            self.dir)
        see, left, right = see_dir(self, self.drop, angle)
        if -math.pi/2 < angle < math.pi/2 and \
                not self.hidden(see_robots, self.drop, angle, left, right):
            if see > 0:
                d = dist_between(self, self.drop)
                self.see_dist.append((d, see, 'drop'))
        #if self.seeing_drop != 0 and see == 0:
        #    self.add_message('not_see(drop)')
        #elif self.seeing_drop != see:
        #    if see != 0:
        #        self.add_message('see(%s, drop)' % dir2text(see))
        #self.seeing_drop = see
        
    def check_seeing(self):
        """Check to see what is seen"""
        see_robots = []
        self.see_dist = []
        # check to see if a robot bottle be seen and 
        # construct the see_robots list to test for hiding
        for robot in self.env.robots:
            if robot == self:
                continue
            see, left, right = see_dir(self, robot)
            if left > -math.pi/2 and right < math.pi/2:
                if see > 0:
                    d = dist_between(self, robot)
                    self.see_dist.append((d, see, 'robot'))
                    see_robots.append((robot, left, right))
        #if self.seeing_robot != 0 and best_see == 0:
        #    self.add_message('not_see(robot)')
        #elif self.seeing_robot != best_see:
        #    if  best_see != 0:
        #        self.add_message('see(%s, robot)' % dir2text(see))

        self.check_seeing_drop(see_robots)

        # check if seeing a bottle
        for bottle in self.env.bottles:
            angle = get_rel_dir(bottle.x - self.x, bottle.y - self.y, self.dir)
            see, left, right = see_dir(self, bottle, angle)
            if -math.pi/2 < angle < math.pi/2 and \
                    not self.hidden(see_robots, bottle, angle, left, right):
                if see > 0:
                    d = dist_between(self, bottle)
                    self.see_dist.append((d, see, 'bottle'))
        #if self.seeing_bottle != 0 and best_see == 0:
        #    self.add_message('not_see(bottle)')
        #elif self.seeing_bottle != best_see:
        #    if best_see != 0:
        #        self.add_message('see(%s, bottle)' % dir2text(see))
        self.see_dist.sort()
        robot_dir = 0
        robot_dist = 1000000
        drop_dir = 0
        drop_dist = 1000000
        bottle_dir = 0
        bottle_dist = 1000000
        self.seeing_drop = 0
        for see_dist, sdir, name in self.see_dist:
            if name == 'robot' and robot_dist == 1000000:
                robot_dist = approx_dist(see_dist)
                robot_dir = dir2text(sdir)
            if name == 'drop' and drop_dist == 1000000:
                drop_dist = approx_dist(see_dist)
                drop_dir = dir2text(sdir)
                self.seeing_drop = sdir
            if name == 'bottle' and bottle_dist == 1000000:
                bottle_dist = approx_dist(see_dist)
                bottle_dir = dir2text(sdir)
   

        #robot_dist_name = dist2name(robot_dist)
        robot_info = (robot_dist, robot_dir)
        if self.saved_robot_see_dist != robot_info:
            if robot_dist ==1000000:
                self.add_message('unset(see(robot, 0, centre))')
            else:
                self.add_message('set(see(robot, {0}, {1}))'.format(robot_dist, robot_dir))
            self.saved_robot_see_dist = robot_info
        #drop_dist_name = dist2name(drop_dist)
        drop_info = (drop_dist, drop_dir)
        if self.saved_drop_see_dist != drop_info:
            if drop_dist ==1000000:
                self.add_message('unset(see(drop, 0, centre))')
            else:
                self.add_message('set(see(drop, {0}, {1}))'.format(drop_dist, drop_dir))
            self.saved_drop_see_dist = drop_info
        #bottle_dist_name = dist2name(bottle_dist)
        bottle_info = (bottle_dist, bottle_dir)
        if self.saved_bottle_see_dist != bottle_info:
            if bottle_dist ==1000000:
                self.add_message('unset(see(bottle, 0, centre))')
            else:
                self.add_message('set(see(bottle, {0}, {1}))'.format(bottle_dist, bottle_dir))
            self.saved_bottle_see_dist = bottle_info

    def hidden(self, see_robots, obj, obj_angle, left, right):
        """Test to see if obj and obj_angle is hidden by a robot"""
        dist2obj = math.sqrt((self.x - obj.x)**2 + (self.y - obj.y)**2)
        rel_left = normalize_angle(left - obj_angle)
        rel_right = normalize_angle(right - obj_angle)
        for robot, rleft, rright in see_robots:
            dist2robot = math.sqrt((self.x - robot.x)**2 + (self.y - robot.y)**2)
            if dist2robot < dist2obj:
                rel_rleft = normalize_angle(rleft - obj_angle)
                rel_rright = normalize_angle(rright - obj_angle)
                if -math.pi/2 < rel_rleft < rel_left and \
                        math.pi/2 > rel_rright > rel_right:
                    return True
        return False

    # for touching
    # self.touching = 0    - not touching
    # OW  or together
    #   1  - touching right
    #   2  - touching left
    #   4  - touching far_right
    #   8  - touching far_left

    def check_touching(self):
        """Determine touching and send as msg"""
        self.touching = 0
        self.check_touching_bottle()
        #self.check_touching_drop()
        self.check_touching_robot()
        self.check_touching_walls()
        if self.touching != self.saved_touching:
            self.add_message('unset(touching(centre))')
            if self.touching == 3:
                self.add_message('set(touching(centre))')
            else:
                if self.touching & 1 == 1:
                    self.add_message('set(touching(right))')
                if self.touching & 2 == 2:
                    self.add_message('set(touching(left))')
                if self.touching & 4 == 4:
                    self.add_message('set(touching(far_right))')
                if self.touching & 8 == 8:
                    self.add_message('set(touching(far_left))')
            self.saved_touching = self.touching

    def check_holding(self):
        if self.holding:
            if not self.last_holding:
                self.add_message('set(holding)')
                self.last_holding = self.holding
        else:
            if self.last_holding:
                self.add_message('unset(holding)')
                self.last_holding = self.holding

    def check_gripper_open(self):
        if self.gripper_open:
            if not self.last_gripper_open:
                self.add_message('set(gripper_open)')
                self.last_gripper_open = self.gripper_open
        else:
            if self.last_gripper_open:
                self.add_message('unset(gripper_open)')
                self.last_gripper_open = self.gripper_open

    def check_touching_bottle(self):
        """Check if touching a bottle and if touching in front return index
        of touching bottle or return -1"""
        index = -1
        for i, bottle in enumerate(self.env.bottles):
            if (self.x - bottle.x)**2 + (self.y -bottle.y)**2 < SQ_R2C + 50:
                angle = get_rel_dir(bottle.x - self.x, bottle.y - self.y, self.dir)
                if -HC_ANGLE/4 < angle < HC_ANGLE/4:
                    self.touching |= 3
                    index = i
                elif HC_ANGLE/4 <= angle < HC_ANGLE:
                    self.touching |= 1
                elif -HC_ANGLE/4 >= angle > -HC_ANGLE:
                    self.touching |= 2
                #elif HC_ANGLE <= angle < math.pi/2:
                #    self.touching |= 4
                #elif -HC_ANGLE >= angle > -math.pi/2:
                #    self.touching |= 8
        return index
    
    def update_touching(self, angle):
        """Update touching given the touching angle"""
        if -HC_ANGLE/4 < angle < HC_ANGLE/4:
            self.touching |= 3
        elif HC_ANGLE/4 <= angle < HC_ANGLE:
            self.touching |= 1
        elif -HC_ANGLE/4 >= angle > -HC_ANGLE:
            self.touching |= 2
        elif HC_ANGLE <= angle < math.pi/2:
            self.touching |= 4
        elif -HC_ANGLE >= angle > -math.pi/2:
            self.touching |= 8

    def check_touching_robot(self):
        """Update touching for robots"""
        for robot in self.env.robots:
            if robot == self:
                continue
            if (self.x - robot.x)**2 + (self.y -robot.y)**2 < SQ_R2R + 100:
                angle = get_rel_dir(robot.x - self.x, 
                                    robot.y - self.y, self.dir)
                self.update_touching(angle)

    def check_touching_drop(self):
        """Update touching for drop."""
        if (self.x - self.drop.x)**2 + (self.y - self.drop.y)**2 < SQ_R2D + 50:
            angle = get_rel_dir(self.drop.x - self.x, 
                                self.drop.y - self.y, self.dir)
            self.update_touching(angle)


    def check_touching_walls(self):
        """Update touching for walls"""
        if self.x < ROBOT_RADIUS+2:
            # touching West wall
            self.update_touching(get_rel_dir(-1,0,self.dir))
        if self.x > WIDTH - ROBOT_RADIUS - 2:
            # touching E
             self.update_touching(get_rel_dir(1,0,self.dir))
        if self.y <  ROBOT_RADIUS+2:
            # touching N
            self.update_touching(get_rel_dir(0,-1,self.dir))
        if self.y > HEIGHT - ROBOT_RADIUS - 2:
            # touching S
             self.update_touching(get_rel_dir(0,1,self.dir))



    def check_radar(self):
        """Determine radar and send as msg"""
        self.radars = [[],[],[],[],[],[],[]]
        self.check_radar_drop()
        self.check_radar_robot()
        self.check_radar_walls()
        self.check_radar_bottles()
        r1 = self.radars[1]
        r2 = self.radars[2]
        r3 = self.radars[3]
        r1.sort()
        r2.sort()
        r3.sort()
        d = min(r1[0], r2[0], r3[0])
        closest = approx_dist(d)
        if closest < 33 and self.saved_radar_right != closest:
            self.add_message('radar_close(right, %d)' % closest)
            self.saved_radar_right = closest
        elif closest > 33 and self.saved_radar_right != -1:
            self.add_message('radar_not_close(right)')
            self.saved_radar_right = -1
        l1 = self.radars[4]
        l2 = self.radars[5]
        l3 = self.radars[6]
        l1.sort()
        l2.sort()
        l3.sort()
        d = min(l1[0], l2[0], l3[0])
        closest = approx_dist(d)
        if closest < 17 and self.saved_radar_left != closest:
            self.add_message('radar_close(left, %d)' % closest)
            self.saved_radar_left = closest
        elif closest > 17 and self.saved_radar_left != -1:
            self.add_message('radar_not_close(left)')
            self.saved_radar_left = -1
        dists = self.radars[0]
        dists.sort()
        closest = approx_dist(dists[0])
        if self.saved_radar_centre != closest:
            self.add_message('radar_centre_dist(%d)' % closest)
            self.saved_radar_centre = closest

    def check_radar_bottles(self):
        """Check if radars see a bottle"""
        for bottle in self.env.bottles:
            dist = get_closest_to(self, bottle, self.dir)
            self.radars[0].append(dist-ROBOT_RADIUS)
    
    def check_radar_robot(self):
        """Update radar for robots"""
        for robot in self.env.robots:
            if robot == self:
                continue
            self.update_radar(robot)

    def update_radar(self, obj):
        for i,ang in enumerate(RADAR_ANGLES):
            dist = get_closest_to(self, obj, self.dir + ang)
            self.radars[i].append(dist-ROBOT_RADIUS)
  
    def check_radar_drop(self):
        """Update radar for drop."""
        drop = self.drop
        self.update_radar(self.drop)

    def check_radar_walls(self):
        """Update radar for walls"""
        for i,ang in enumerate(RADAR_ANGLES):        
            dist = get_closest_wall(self, self.dir + ang)
            self.radars[i].append(dist-ROBOT_RADIUS)


    def check_over_drop(self):
        x,y = self.drop.x, self.drop.y
        over = (self.x - x)**2 + (self.y - y)**2 < SQ_R2D
        if over:
            if not self.over:
                self.add_message('set(over_drop)')
                self.over = True
        else:
            if self.over:
                self.add_message('unset(over_drop)')
                self.over = False


    def check_close_drop(self):
        angle = get_rel_dir(self.drop.x - self.x, 
                            self.drop.y - self.y, self.dir)
        close = -HC_ANGLE < angle < HC_ANGLE and \
            (self.x - self.drop.x)**2 + (self.y - self.drop.y)**2 < C_SQ_R2D
        if close:
            if not self.close_drop:
                self.add_message('set(close(drop))')
                self.close_drop = True
        else:
            if self.close_drop:
                self.add_message('unset(close(drop))')
                self.close_drop = False
 

    def check_close_robot(self):
        """Check if close to a robot"""
        close = False
        for robot in self.env.robots:
            if robot == self:
                continue
            angle = get_rel_dir(robot.x - self.x, robot.y - self.y, self.dir)
            if  -HC_ANGLE < angle < HC_ANGLE \
                    and (self.x - robot.x)**2 + (self.y -robot.y)**2 < C_SQ_R2R:
                # in front and close
                close = True
                break
        if close:
            if not self.close_robot:
                self.add_message('set(close(robot)(')
                self.close_robot = True
        else:
            if self.close_robot:
                self.add_message('unset(close(robot))')
                self.close_robot = False
            
    def check_close_bottle(self):
        """Check if code to bottle - must be in view angle."""
        index = -1
        for i in range(len(self.env.bottles)):
            bottle = self.env.bottles[i]
            angle = get_rel_dir(bottle.x - self.x, bottle.y - self.y, self.dir)
            if -HC_ANGLE < angle < HC_ANGLE \
                   and (self.x - bottle.x)**2 + (self.y -bottle.y)**2 < C_SQ_R2C:
                    index = i
                    break

        if index == -1:
            if self.close_bottle:
                self.add_message('unset(close(bottle))')
                self.close_bottle = False
        else:
            if not self.close_bottle:
                self.add_message('set(close(bottle))')
                self.close_bottle = True

            
    def real_dir(self):
        if self.speed < 0:
            return -self.dir
        else:
            return self.dir

    def update_speed(self):
        """Update the robot velocity based on touching walls, drop or other robot"""
        # modify angle if turning - keep dir in [0,2*pi]
        if self.turning != 0:
            self.dir += self.turning*0.04
            if self.dir < 0: self.dir += 2*math.pi
            if self.dir > 2*math.pi: self.dir -= 2*math.pi
        # only bother if robot is moving
        if self.moving or self.percept_update:
            self.percept_update = False
            # self the x and y velocity - modified by collisions
            self.vx = self.speed*math.cos(self.dir)
            self.vy = self.speed*math.sin(self.dir)
            #print self.speed, self.vx, self.vy
            # Am I colliding with walls?
            touching_wall = False
            if self.x < ROBOT_RADIUS+2 and self.vx < 0: 
                touching_wall = True
                self.vx = 0
            if self.x > WIDTH - ROBOT_RADIUS-2 and self.vx > 0: 
                touching_wall = True
                self.vx = 0
            if self.y < ROBOT_RADIUS+2 and self.vy < 0:
                touching_wall = True
                self.vy = 0
            if self.y > HEIGHT - ROBOT_RADIUS-2 and self.vy > 0:
                touching_wall = True
                self.vy = 0
            # Am I colliding with the drop
            if self.vx == 0 and self.vy == 0:
                return
            robot_dir = get_dir(self.vx, self.vy)
            """
            rel_drop_dir = get_rel_dir(self.drop.x-self.x, self.drop.y-self.y, robot_dir)
            if (self.x - self.drop.x)**2 + (self.y - self.drop.y)**2 < SQ_R2D + 4 and -math.pi/2 < rel_drop_dir < math.pi/2:
                self.vx, self.vy = 0,0
            """
            # Am I colliding with another robot
            if abs(self.vx) + abs(self.vy) < 0.00001:  # used to be two .vx
                self.vx, self.vy = 0,0
                return
            for robot in self.env.robots:
                if robot == self:
                    continue
                robot_dir = get_rel_dir(robot.x - self.x, robot.y - self.y, self.dir)
                if (robot.x - self.x)**2 + (robot.y - self.y)**2 < SQ_R2R + 8 and -math.pi/2 < robot_dir < math.pi/2:
                    robot_dir = get_dir(robot.x - self.x, robot.y - self.y)
                    speed = -self.vx*math.sin(robot_dir) + \
                        self.vy*math.cos(robot_dir)
                    self.vx = -speed*math.sin(robot_dir)
                    self.vy = speed*math.cos(robot_dir)
                    if touching_wall:
                        self.vx, self.vy = 0, 0
                    break
        else:
            self.vx, self.vy = 0, 0
        

    def update(self):
        """Update the robot position and possibly move bottles"""
        # only bother if robot is moving
        if self.moving:
            if self.veering:
                self.veer_steps -= 1
                if self.veer_steps <= 0:
                    self.veering = False
                    self.stop()
                    self.add_message('end_veering')
                elif self.veer_steps == VEER_STEPS/2:
                    self.turning *= -1
            # check bottles to see if they need to move out of the way
            dir = get_dir(self.vx, self.vy)
            for i in range(len(self.env.bottles)):
                bottle = self.env.bottles[i]
                cx = bottle.x
                cy = bottle.y
                if (self.x - cx)**2 + (self.y -cy)**2 < SQ_R2C + 4:
                    # robot is touching bottle
                    angle = get_dir(cx - self.x, cy - self.y)
                    diff = get_rel_dir(cx - self.x, cy - self.y, dir)
                    if abs(diff) < math.pi/2 or abs(diff) > 3*math.pi/2:
                        # the bottle is in the front half
                        # TODO - robot pushes bottle into something
                        if abs(diff) < 0.2:
                            # the bottle is straight in front so move with robot
                            dx = self.speed*math.cos(dir)
                            dy = self.speed*math.sin(dir)
                            self.env.bottles[i].move_delta(dx, dy)
                        else:
                            # bottle moves around robot (tangent)
                            cspeed = self.speed*math.cos(diff)
                            self.env.bottles[i].move_delta(cspeed*math.cos(angle),
                                cspeed*math.sin(angle))
            
            self.x += self.vx
            self.y += self.vy
            #print self.x, self.y
        # now that things have changed recheck
        self.check_seeing()
        self.check_touching()
        self.check_over_drop()
        #self.check_radar()
        #self.check_close_drop()
        #self.check_close_bottle()
        #self.check_close_robot()
        self.check_holding()
        self.check_gripper_open()
        self.send_messages()


    def move_forward(self, speed):
        #print 'move_forward', speed
        self.speed = speed*SPEED_FACTOR
        self.moving = True
        
    def turn(self, tdir, rate):
        if tdir == 'left' or tdir == 'far_left':
            self.turning = -rate*RATE_FACTOR
        else:
            self.turning = rate*RATE_FACTOR


    def start_veering(self, dir):
        self.turn(dir, 0.1)
        self.move_forward(1.2)
        self.veer_steps = VEER_STEPS
        self.veering = True

    def stop(self):
        self.moving = False
        self.turning = 0
        if  self.veering:
            self.add_message('end_veering')
            self.veering = False
        
    def stop_turn(self):
        self.turning = 0

    def stop_move(self):
        self.moving = False
        if  self.veering:
            self.add_message('end_veering')
            self.veering = False
        
    def grip(self):
        self.gripper_open = False
        index = self.check_touching_bottle()
        if index == -1 : return
        self.holding = True
        self.env.bottles.pop(index)
        self.percept_update = True
#        self.send_msg('holding')

    def release(self):
        self.gripper_open = True
        if self.holding: # and (self.touching & 1) and self.seeing_drop == 1:
            self.holding = False
            if USING_SOUND:
                self.env.parent.ting_sound.play()
            d = self.dir
            x = self.x + math.cos(d)*(ROBOT_RADIUS + 2*BOTTLE_RADIUS)
            y = self.y + math.sin(d)*(ROBOT_RADIUS + 2*BOTTLE_RADIUS)
            self.dropped_can = self.env.addBottle(x,y)
            self.drop_timer.Start(DROP_TIME, oneShot=True)
#            self.percept_update = True
#            self.send_msg('not_holding')
#        else:
#            print self.holding, self.touching, self.seeing_drop
            
    def unset_holding(self):
#        self.send_msg('not_holding')
#        self.percept_update = True
        self.holding = False
        
class Bottle:
    def __init__(self, env, x, y):
        self.x = x
        self.y = y
        self.ID = env.inc_ID()
        self.radius = BOTTLE_RADIUS

    def move_delta(self, dx, dy):
        self.x += dx
        self.y += dy

    def move(self, x, y):
        self.x = x
        self.y = y

class Environment:
    def __init__(self, parent):
        self.parent = parent

        self.bottles = []
        self.bottleID = 0
        self.leftdown= False
        self.messages_to_send = None
        robot0 = Robot(self, 1, WIDTH*0.25, HEIGHT/2)
        if NUM_ROBOTS == 1:
            self.robots = [robot0]
        else:
            robot1 = Robot(self, 2, WIDTH*0.75, HEIGHT/2)
            self.robots = [robot0, robot1]

    def inc_ID(self):
        self.bottleID += 1
        return self.bottleID

    def flipStopped(self):
        for r in self.robots:
            r.flipStopped()
            
    def bottle_place(self, x, y):
        dx,dy = DROP_POSITION
        if x > dx - DROP_RADIUS and y <  dy + DROP_RADIUS:
            return False
        return ROBOT_RADIUS < x < WIDTH - ROBOT_RADIUS and \
            ROBOT_RADIUS < y < HEIGHT - ROBOT_RADIUS

    def addBottle(self, x, y):
        bottle = Bottle(self, x,y)
        self.bottles.append(bottle)
        return bottle.ID

    def getBottle(self, x, y):
        for i in range(len(self.bottles)):
            bottle = self.bottles[i]
            if abs(bottle.x - x) < BOTTLE_RADIUS and abs(bottle.y - y) < BOTTLE_RADIUS:
                return bottle
        return None

    def removeBottleID(self, BID):
        for b in self.bottles:
            if b.ID == BID:
                self.bottles.remove(b)
                for robot in self.robots:
                    robot.check_seeing()
                    robot.check_touching()
                    robot.check_radar()
                break

    def removeBottle(self, x, y):
        for i in range(len(self.bottles)):
            bottle = self.bottles[i]
            if abs(bottle.x - x) < BOTTLE_RADIUS and abs(bottle.y - y) < BOTTLE_RADIUS:
                self.bottles.pop(i)
                for robot in self.robots:
                    robot.check_seeing()
                    robot.check_touching()
                    robot.check_radar()
                break
        for robot in self.robots:
            if robot.holding and  \
                    abs(robot.x - x) < ROBOT_RADIUS and \
                    abs(robot.y - y) < ROBOT_RADIUS:
                robot.holding = False
#                self.percept_update = True
#                robot.send_msg('not_holding')
        """
        elif event.button == 1:
        for bottle in self.bottles:
        rect = Rect(bottle.x-BOTTLE_RADIUS, bottle.y-BOTTLE_RADIUS, 
        2*BOTTLE_RADIUS, 2*BOTTLE_RADIUS)
        px,py = event.pos
        if rect.collidepoint(px,py):
        self.leftdown = True
        break
        """

    def step(self):
        for r in self.robots:
            if r.started:
                r.update_speed()
                r.update()

 

    def process_msg(self, term):
        msg = term.args[2]
        sender_process = term.args[1].args[0].args[1].val
        robotID = int(sender_process[-1])-1
        #print msg
        if str(msg) == "initialise_":
            robotID = int(sender_process[-1])-1
            self.robots[robotID].started = True
            return
        elif  msg.get_type() == pedroclient.PObject.atomtype and \
                msg.val == 'finish':
            return
        elif  msg.get_type() == pedroclient.PObject.structtype and \
                msg.functor.val ==  'msg':
            cmd = msg.args[0].functor.val
            if not TWO_TASK:
                self.robots[robotID].message = str(msg.args[0])
                if (MESSAGE_TIME > 0):
                    self.robots[robotID].message_timer.Stop()
                    self.robots[robotID].message_timer.Start(MESSAGE_TIME, 
                                                             oneShot=True)
            if cmd == 'count':
                cmd_args = msg.args[0].args
                if NUM_ROBOTS == 1:
                    self.robots[0].count = cmd_args[0].val
                else:
                    self.robots[1-robotID].count = cmd_args[0].val
 

        elif  msg.get_type() == pedroclient.PObject.structtype and \
                msg.functor.val ==  'controls':
            actions = msg.args[0]
            if actions.get_type() != pedroclient.PObject.listtype:
                print "COMMAND: ", str(pmsg)
                assert False
            for a in actions.toList():
                self.process_action(a, sender_process, robotID)
            return
  
    def process_action(self, msg, sender_process, robotID):
        if TWO_TASK:
            assert msg.get_type() == pedroclient.PObject.structtype
            functor = msg.functor
            cmd_type = functor.val
            cmd = msg.args[0]
            assert cmd.get_type() == pedroclient.PObject.structtype
            robotID = int(cmd.args[0].val[-1]) - 1
            cmd_args = cmd.args[1:]
            cmd = cmd.functor.val
            #print cmd_type, cmd, str(cmd_args)
        else:
            #sender_process = term.args[1].args[0].args[1].val
            robotID = int(sender_process[-1]) - 1
            assert msg.get_type() == pedroclient.PObject.structtype
            functor = msg.functor
            assert functor.get_type() == pedroclient.PObject.atomtype
            cmd_type = functor.val
            full_cmd = msg.args[0]
            if full_cmd.get_type() == pedroclient.PObject.atomtype:
                cmd_args = []
                cmd = full_cmd.val
            else:
                assert full_cmd.get_type() == pedroclient.PObject.structtype
                cmd_args = full_cmd.args
                cmd = full_cmd.functor.val
        if cmd_type == 'stop_':
            if cmd in ['drop', 'hold']:
                return
            elif cmd == 'turn':
                self.robots[robotID].stop_turn()
            elif cmd == 'move':
                self.robots[robotID].stop_move()
            else:
                print "COMMAND: ", str(msg)
                assert False
        elif cmd_type in ['start_', 'mod_']:
            if cmd == 'turn':
                self.robots[robotID].turn(cmd_args[0].val, cmd_args[1].val)
            elif cmd == 'move':
                self.robots[robotID].move_forward(cmd_args[0].val)
            else:
                print "COMMAND: ", cmd, str(msg)
                assert False
        elif cmd_type == 'exec_':
            if cmd == 'close_gripper':
                self.robots[robotID].grip()
            elif cmd == 'open_gripper':
                self.robots[robotID].release()
                return
            else:
                print "COMMAND: ", str(msg)
                assert False
        else:
            print "COMMAND: ", str(msg)
            assert False
 
    def send_msg(self, ID, msg):
        if TWO_TASK:
            self.parent.client.p2p("percepts:robots", msg)
            """
            if self.messages_to_send is None:
                self.messages_to_send = msg
            else:
                if msg == '[]':
                    all_messages = self.messages_to_send
                elif self.messages_to_send == '[]':
                    all_messages = msg
                else:
                    all_messages = self.messages_to_send + msg
                    all_messages = all_messages.replace('][', ',')
                self.parent.client.p2p("percepts:robots", all_messages)
                self.messages_to_send = None
                """
        else:
            self.parent.client.p2p("percepts:robot%d" % ID, msg)
            #print "sending message to", "percepts:robot%d" % ID, msg

    def send_message(self, ID, msg):
        if TWO_TASK:
             self.parent.client.p2p("messages:robots",  msg)
        else:
            self.parent.client.p2p("messages:robot%d" % ID, msg)
        #print "sending message to", "messages:robot%d" % ID, msg

    def send_log(self, msg):
        self.parent.client.p2p("thread0:logger2", '['+msg+',nl]')

class BottlesApp(wx.App):
    """Bottles Visualisation"""

    def OnInit(self):
        controller = Controller("Bottles", (WIDTH, HEIGHT+TEXT_AREA_HEIGHT))
        controller.SetClientSize((WIDTH, HEIGHT+TEXT_AREA_HEIGHT))
        controller.Bind(wx.EVT_CLOSE, self.onClose)
        controller.Show()
        self.SetTopWindow(controller)
        self.controller = controller
        return True

    def onClose(self, evt):
        if TWO_TASK:
            self.controller.env.send_message(1, "quiting")
        else:
            self.controller.env.send_message(1, "quiting")
            self.controller.env.send_message(2, "quiting")
        time.sleep(1)
        self.controller.Destroy()

class View:
    def __init__(self, parent):
        self.parent = parent
        self.white_pen = wx.Pen(wx.WHITE)
        self.white_brush = wx.Brush(wx.WHITE)
        self.black_pen = wx.Pen(wx.BLACK)
        self.black_brush = wx.Brush(wx.BLACK)
        self.bottle_colour_pen = wx.Pen(BOTTLE_COLOUR)
        self.bottle_colour_brush= wx.Brush(BOTTLE_COLOUR)
        self.red_pen = wx.Pen(wx.RED)
        self.red_brush = wx.Brush(wx.RED)
        self.robot_colour_pen = wx.Pen(ROBOT_COLOUR)
        self.robot_colour_brush = wx.Brush(ROBOT_COLOUR)

    def redraw(self):
        dc = wx.PaintDC(self.parent.panel)
        #font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        #font.SetStyle(wx.FONTWEIGHT_BOLD)
        #font.SetPointSize(FONT_SIZE)
        font = wx.Font(FONT_SIZE, wx.MODERN, wx.NORMAL, wx.BOLD, 
                       False, FONT_NAME)
        dc.SetFont(font)
        dc.SetPen(self.white_pen)
        dc.SetBrush(self.white_brush)
        dc.DrawRectangle(0, HEIGHT, WIDTH, HEIGHT + TEXT_AREA_HEIGHT)
        dc.SetPen(self.red_pen)
        dc.SetBrush(self.red_brush)
        x,y = DROP_POSITION
        dc.DrawCircle(x,y,DROP_RADIUS)
        for robot in self.parent.getRobots():
            self.drawRobot(dc, robot)
        dc.SetPen(self.bottle_colour_pen)
        dc.SetBrush(self.bottle_colour_brush)
        for bottle in self.parent.getBottles():
            dc.DrawCircle(int(bottle.x), int(bottle.y), BOTTLE_RADIUS)
            

    def drawRobot(self, dc, robot):
        """Draw the robot"""
        colour = ROBOT_COLOUR
        x,y = robot.x, robot.y
        d = robot.dir
        dc.SetPen(self.robot_colour_pen)
        dc.SetBrush(self.robot_colour_brush)
        dc.DrawCircle(int(x),int(y), ROBOT_RADIUS)
        x1,y1 = x + ROBOT_RADIUS*math.cos(d - HC_ANGLE), \
            y + ROBOT_RADIUS*math.sin(d - HC_ANGLE)
        x2,y2 = x + ROBOT_RADIUS*math.cos(d + HC_ANGLE), \
            y + ROBOT_RADIUS*math.sin(d + HC_ANGLE)
        dc.SetPen(wx.Pen(CAMERA_ARC_COLOUR))
        dc.SetBrush(wx.Brush(CAMERA_ARC_COLOUR))
        dc.DrawArc(x2, y2, x1, y1, x, y)
        wheel1 = [(15,14),(15,10),(7,10),(7,14)]
        wheel2 = [(15,-14),(15,-10),(7,-10),(7,-14)]
        transform(x, y, d, wheel1)
        transform(x, y, d, wheel2)
        dc.SetPen(self.black_pen)
        dc.SetBrush(self.black_brush)
        dc.DrawPolygon(wheel1)
        dc.DrawPolygon(wheel2)
        backwheel = [(-16, 0)]
        transform(x, y, d, backwheel)
        cx,cy = backwheel[0]
        dc.DrawCircle(int(cx),int(cy), 4)
        if robot.ID == 1:
            colour = ROBOT1_COLOUR
            dc.SetTextForeground(ROBOT1_COLOUR_TRIPLE)
        else:
            colour = ROBOT2_COLOUR
            dc.SetTextForeground(ROBOT2_COLOUR_TRIPLE)
        dc.SetPen(wx.Pen(colour))
        dc.SetBrush(wx.Brush(colour))
        dc.DrawCircle(x, y, ROBOT_ID_RADIUS)
        if TWO_TASK:
            if robot.ID == 1:
                dc.DrawText(TWO_DISPLAY_TEXT.format(robot.count),
                    10+(robot.ID - 1)*WIDTH/2, HEIGHT + 10)
        elif NUM_ROBOTS == 1:    
            dc.DrawText(ONE_DISPLAY_TEXT.format(robot.count),
                        10+(robot.ID - 1)*WIDTH/2, HEIGHT + 10)
        else:    
            dc.DrawText(DISPLAY_TEXT.format(robot.count, robot.message),
                        10, HEIGHT + 30*robot.ID - 30)
        if robot.holding:
            bottle_pos = [(16, 0)]
            transform(x, y, d, bottle_pos)
            cx,cy = bottle_pos[0]
            dc.SetPen(self.bottle_colour_pen)
            dc.SetBrush(self.bottle_colour_brush)
            dc.DrawCircle(int(cx),int(cy), BOTTLE_RADIUS) 
        

class Controller(wx.Frame):
    def __init__(self, title, s):
        wx.Frame.__init__(self, None, -1, title, 
                          style=wx.DEFAULT_FRAME_STYLE & ~wx.RESIZE_BORDER)
        #                  size = s)
        # Graphics area
        self.panel = wx.Panel(self)
        self.panel.Bind(wx.EVT_LEFT_DOWN, self.button_press_cb)
        self.panel.Bind(wx.EVT_LEFT_UP, self.button_release_cb)
        self.panel.Bind(wx.EVT_MIDDLE_DOWN, self.mid_button_press_cb)
        self.panel.Bind(wx.EVT_RIGHT_DOWN, self.right_button_press_cb)
        self.panel.Bind(wx.EVT_MOTION, self.motion_cb)
        self.panel.Bind(wx.EVT_KEY_DOWN, self.onKeyPress)
        self.panel.SetFocus()
        self.running = True
        self.env = Environment(self)
        self.view = View(self)
        self.client = pedroclient.PedroClient()
        self.client.register('env')
        self.thread = GetNotifications(self.client, self.panel)
        self.thread.setDaemon(True)
        self.thread.start()
        # Callback for notification event
        self.panel.Bind(EVT_NOTIFICATION_EVENT, self.data_cb)
        self.timer = wx.Timer(self, TIMER_ID)
        wx.EVT_TIMER(self, TIMER_ID, self.timer_cb)
        self.timer.Start(30, False)
        self.panel.Bind(wx.EVT_PAINT, self.area_expose_cb)
        self.moving_bottle = None
        #self.sound = wx.Sound('ting.wav')
        #self.sound.Play(wx.SOUND_ASYNC)
        if USING_SOUND:
            mixer.init()
            self.ting_sound = mixer.Sound(TING_FILE)
            self.swoosh_sound = mixer.Sound(SWOOSH_FILE)
         # self.ting_sound.play()

    def onKeyPress(self, event):
        if event.GetKeyCode() != 32:
            return
        dc = wx.PaintDC(self.panel)
        dc.SetTextForeground('#FF8800') 
        dc.DrawText( "STOPPED", 350,550)
        self.env.flipStopped()
        self.running = not self.running

    def timer_cb(self, event):
        if self.running:
            self.env.step()
            self.Refresh()

    def data_cb(self, event):
        self.env.process_msg(event.notification)

    def button_press_cb(self, event):
        # Keith
        if event.AltDown() and event.CmdDown():
            self.env.removeBottle(event.GetX(), event.GetY())
        elif event.AltDown():
            self.env.addBottle(event.GetX(), event.GetY())
        else:
            self.moving_bottle = self.env.getBottle(event.GetX(), event.GetY())

    def button_release_cb(self, event):
        self.moving_bottle = None

    def motion_cb(self, event):
        if self.moving_bottle is not None:
            self.moving_bottle.move(event.GetX(), event.GetY())

    def mid_button_press_cb(self, event):
        self.env.addBottle(event.GetX(), event.GetY())

    def right_button_press_cb(self, event):
        self.env.removeBottle(event.GetX(), event.GetY())

    def area_expose_cb(self, e):
        self.view.redraw()

    def getRobots(self):
        return self.env.robots

    def getBottles(self):
        return self.env.bottles

if __name__ == "__main__":
    if len(sys.argv) == 2:
        if sys.argv[1].lower() == "twotasks":
            TWO_TASK = True
        elif sys.argv[1].lower() == "one":
            NUM_ROBOTS = 1
    bottles_app = BottlesApp()
    bottles_app.MainLoop()
